// STL Exporter for Lightwave 7.0+
//
//   Written By Sean Hyde-Moyer (Copywrite 2003)
//   This program is released to the public under the terms of the
//   GNU GENERAL PUBLIC LICENSE
//   See http://www.gnu.org/copyleft/gpl.txt
//   For the full text of the license agreement



@script modeler


main
{

	fType = 1;
	fDefault[1] = "Binary";
	fDefault[2] = "ASCII";

	pType = 1;
	pDefault[1] = "Intel";
	pDefault[2] = "Mac";


//______Prep File to Write_______________________________________________________________


    	directory = recall ("directory","Drive");



	reqbegin("SHM STL Exporter");
    	reqsize(500,195);

    	YPOS = 10; 

		w1 = ctltext("","Be sure to save you object before running the exporter. Some alterations may occur.");
        		ctlposition(w1,10,YPOS,450,20);
			YPOS = YPOS + 21;

		w2 = ctltext("","For best results, object should reside on a single layer.");
        		ctlposition(w2,10,YPOS,450,20);
			YPOS = YPOS + 30;

		f1 = ctlfilename("Save File As",directory,100,0);
        		ctlposition(f1,10,YPOS,450,20);
			YPOS = YPOS + 25;

 		t1 = ctlchoice("Type",fType,fDefault);
        		ctlposition(t1,71,YPOS,124,20);
			YPOS = YPOS + 25;

 		t2 = ctlchoice("Byte Order",pType,pDefault);
        		ctlposition(t2,45,YPOS,150,20);
			YPOS = YPOS + 25;

		s1 = ctlcheckbox("Fix Inverted", true);
        		ctlposition(s1,100,YPOS,150,20);

	if(reqpost())
	{
		stlFileName = getvalue(f1);
		writeType = getvalue(t1);
		platformType = getvalue(t2);
		invertIt = getvalue(s1);
	}
	else
		return;
	reqend();

	isPC = TRUE;
	if(platformType == 2) isPC = FALSE;


	//_____________________Do Some Error Checking


	if(stlFileName == nil)
		return;


	base = split(stlFileName);

	stlname = base[3];



	//_____________________Make the extension STL


	newFilename = string(base[1],base[2],base[3],".stl");

	if(fileexists(newFilename)){

		reqbegin("SHM STL Exporter");
    		reqsize(300,80);

    		YPOS = 10; 

		w1 = ctltext("Warning! File Exists.","Press OK to overwrite, or CANCEL to quit.");
        		ctlposition(w1,30,YPOS,280,20);
		if(reqpost())
		{

		}
		else
			return;
		reqend();

	}


	if(writeType == 1){
		stlFile = File(newFilename,"wb");
	}
	else{
		stlFile = File(newFilename,"w");
	}


	if(stlFile == nil){
		error("Cannot Create File \"",stlFileName,"\"");
		return;
	}

        store("directory",stlFileName);



//______Prep File For STL Compatibility


	selmode(USER);
	selpolygon(CLEAR,NVGT,0); // Clear User Selections

	mergepoints(0);  // Merge Overlapping Points

	triple();  // Triple all polygons

	testbadpolys = polycount(); // Delete 1 & 2 point polygons
	if((testbadpolys[2] + testbadpolys[3]) > 0)
	{

		selmode(USER);
		selpolygon(SET,NVLT,3);
		cut();                  
	}

	if(invertIt == 1){
		scale(<-1,1,1>);
		flip();
	}

	positively(); // Move so all vertices are represented by positive numbers, as per STL Format Definition
	



//______Export as Binary STL_____________________________________________________________________


	if(writeType == 1){

		//______Write Header of 80 Bytes; Content is Irrelevant__________________________


		headerString = string(stlname," SHM STL Exporter for Lightwave 3D");	

		for(headloop = 1; headloop <= 80; ++headloop)
		{
			if(size(headerString) >= headloop){

				currentChar = ascii(strsub(headerString,headloop,1));
				stlFile.writeByte(currentChar);
			}
			else{
				stlFile.writeByte(0);				
			}
		}	 

		selmode(USER);				
		selpolygon(CLEAR,NVGT,0);

		editbegin();
		polynfo = polycount();



		//______Write Polycount (4 Byte Int)______________________________________________

	
		writeInt(stlFile,polynfo[1],isPC);
	
		moninit(polynfo[1],"Exporting Binary STL");
	
	
		//______Write Body of File________________________________________________________


		for(loop=1; loop <= polynfo[1]; loop++)	
		{
		
			monstep();

			//______Write Normal as 3 Floats__________________________________________

	
			facetNormal = string(polynormal(polygons[loop]));

			normArray = parse(" ",facetNormal);

				for(normloop = 1; normloop <=3; ++normloop)
				{

					tempNorm = number(normArray[normloop]);

					writeFloat(stlFile,tempNorm,isPC);
					
		
				}			


				//______Write Vertices as 3 Floats________________________________
	

				pointarray = polyinfo(polygons[loop]);
	
				currentpolypoint[1] = string(pointinfo(pointarray[2]));
				currentpolypoint[2] = string(pointinfo(pointarray[3]));
				currentpolypoint[3] = string(pointinfo(pointarray[4]));
	
				for(vertloop = 1; vertloop <=3; ++vertloop)
				{

					vertArray = parse(" ",currentpolypoint[vertloop]);

					for(axloop = 1; axloop <=3; ++axloop)
					{
						tempVert = number(vertArray[axloop]);

						writeFloat(stlFile,tempVert,isPC);
												
					}
				}

				//______Write 2 Byte Pad (Could Contain Color Extension)__________________________________________
			
					
				stlFile.writeByte(0);
				stlFile.writeByte(0);			

		}


		editend();
		monend();

	stlFile.close();

	}



//____________Export as Ascii STL_______________________________________________________________


	if(writeType == 2){

		stlFile.writeln("solid ",stlname);  //_Write Header


			selmode(GLOBAL);				
			selpolygon(CLEAR,NVGT,0);

			editbegin();
			polynfo = polycount();

			moninit(polynfo[1],"Exporting Binary STL");	

			for(loop=1; loop <= polynfo[1]; loop++)	
			{
			
				monstep();

				facetNormal = string(polynormal(polygons[loop]));

				stlFile.writeln("  facet normal ",facetNormal);

				stlFile.writeln("   outer loop");
		
					pointarray = polyinfo(polygons[loop]);
		
					currentpolypoint[1] = string(pointinfo(pointarray[2]));
					currentpolypoint[2] = string(pointinfo(pointarray[3]));
					currentpolypoint[3] = string(pointinfo(pointarray[4]));
		
					for(vertloop = 1; vertloop <=3; ++vertloop)
					{
						stlFile.writeln("    vertex ",currentpolypoint[vertloop]);
					}

				stlFile.writeln("   endloop");

				stlFile.writeln("  endfacet");

			}


			editend();
			monend();

		stlFile.writeln("endsolid ",stlname);  //_Write Closer

	}

	if(invertIt ==1){
		undoIt = 6;
	}
	else{
		undoIt = 4;
	}

	for(undoLoop =1; undoLoop <= undoIt; ++undoLoop){
		undo();
	}
}



//_____USER DEFINED FUNCTIONS

positively
{

	//Moves object into all positive cartesian coordinates


	//_____Rests Object @ X+_________________________

    	bbox = (boundingbox());
    	lovecarray = parse(" ",string(bbox[1]));
    	invrse = - number(lovecarray[1]);
    	move(invrse,0,0);

	//_____Rests Object @ Y+_________________________

    	bbox = (boundingbox());
    	lovecarray = parse(" ",string(bbox[1]));
    	invrse = - number(lovecarray[2]);
    	move(0,invrse,0);

	//_____Rests Object @ Z+_________________________

    	bbox = (boundingbox());
    	lovecarray = parse(" ",string(bbox[1]));
    	invrse = - number(lovecarray[3]);
    	move(0,0,invrse);


return;
}






/*
======================================================================
Binary number read/write functions
Ernie Wright  15 Oct 03
====================================================================== */


/*
======================================================================
readShort()    2-byte signed two's-complement integers
readInt()      4-byte signed two's-complement integers
readFloat()    4-byte IEEE floating-point

Read a number from a binary file.  If 'pcOrder' is TRUE, the number is
read in little-endian (Windows) byte order, otherwise it's read in
big-endian (Motorola, Internet) order.
====================================================================== */

readShort : inFile, pcOrder
{
   var byte[ 2 ];
   
   readBytes( inFile, byte, 2, pcOrder );

   if ( byte[ 1 ] < 128 )
      return byte[ 1 ] * 256 + byte[ 2 ];
   else
      return -(( 255 - byte[ 1 ] ) * 256 + ( 256 - byte[ 2 ] ));
}

   
readInt : inFile, pcOrder
{
   var byte[ 4 ];
   
   readBytes( inFile, byte, 4, pcOrder );

   if ( byte[ 1 ] < 128 )      
      return byte[ 1 ] * 16777216
           + byte[ 2 ] * 65536
           + byte[ 3 ] * 256
           + byte[ 4 ];
   else
      return -(( 255 - byte[ 1 ] ) * 16777216
             + ( 255 - byte[ 2 ] ) * 65536
             + ( 255 - byte[ 3 ] ) * 256
             + ( 256 - byte[ 4 ] ));
}


readFloat : inFile, pcOrder
{
   var byte[ 4 ];
   
   readBytes( inFile, byte, 4, pcOrder );

   if ( byte[ 1 ] == 0 && byte[ 2 ] == 0 && byte[ 3 ] == 0 && byte[ 4 ] == 0 )
	return 0;

   m = 1.0 + (( byte[ 2 ] % 128 ) * 65536 + byte[ 3 ] * 256 + byte[ 4 ] )
     / 8388608.0;
   e = ( byte[ 1 ] % 128 ) * 2 + integer( byte[ 2 ] / 128 ) - 127;
   x = m * pow( 2.0, e );
   if ( byte[ 1 ] > 127 ) x = -x;

   return x;
}


/*
======================================================================
writeShort()    2-byte signed two's complement integers
writeInt()      4-byte signed two's complement integers
writeFloat()    4-byte IEEE floating-point

Write a number to a binary file.  If 'pcOrder' is TRUE, the number is
written in little-endian (Windows) byte order, otherwise it's written
in big-endian (Motorola, Internet) order.
====================================================================== */

writeShort : outFile, i, pcOrder
{
   var byte[ 2 ];

   if ( i >= 0 ) {
      byte[ 1 ] = integer( i / 256 ) % 256;
      byte[ 2 ] = i % 256;
   }
   else {
      j = -i - 1;
      byte[ 1 ] = 255 - ( integer( j / 256 ) % 256 );
      byte[ 2 ] = 255 - ( j % 256 );
   }
   
   writeBytes( outFile, byte, 2, pcOrder );
} 


writeInt : outFile, i, pcOrder
{
   var byte[ 4 ];

   if ( i >= 0 ) {
      byte[ 1 ] = integer( i / 16777216 );
      byte[ 2 ] = integer( i / 65536 ) % 256;
      byte[ 3 ] = integer( i / 256 ) % 256;
      byte[ 4 ] = i % 256;
   }
   else {
      j = -i - 1;
      byte[ 1 ] = 255 - integer( j / 16777216 );
      byte[ 2 ] = 255 - ( integer( j / 65536 ) % 256 );
      byte[ 3 ] = 255 - ( integer( j / 256 ) % 256 );
      byte[ 4 ] = 255 - ( j % 256 );
   }
   
   writeBytes( outFile, byte, 4, pcOrder );
}


writeFloat : outFile, x, pcOrder
{
   var byte[ 4 ];

   if(x==0)
   {
      	byte[ 4 ] = 0;
	byte[ 3 ] = 0;
	byte[ 2 ] = 0;
	byte[ 1 ] = 0;
   }
   else
   {
	   lgx = log( abs( x )) / log( 2.0 );
	   e = integer( floor( lgx ));
	   m = integer( floor(( pow( 2.0, lgx - e ) - 1.0 ) * 8388608 ));
	   e += 127;

	   byte[ 1 ] = integer( e / 2 ) + ( x < 0 ? 128 : 0 );
	   byte[ 2 ] = integer( m / 65536 ) + 128 * ( e % 2 );
	   byte[ 3 ] = integer( m / 256 ) % 256;
	   byte[ 4 ] = m % 256;
   }

   writeBytes( outFile, byte, 4, pcOrder );

}


/*
======================================================================
readBytes()

Read bytes from a file.  Called by the number reading functions.

   inFile      the file to read
   byte        array where bytes are stored
   n           the number of bytes to read from inFile
   pcOrder     bytes are stored in reverse order if TRUE
====================================================================== */

readBytes : inFile, byte, n, pcOrder
{
   if ( pcOrder ) {
      first = n;
      stop  = 0;
      step  = -1;
   }
   else {
      first = 1;
      stop  = n + 1;
      step  = 1;
   }
   
   for ( i = first; i != stop; i += step )
      byte[ i ] = inFile.readByte();
}


/*
======================================================================
writeBytes()

Write an array of bytes to a file.  Called by the number writing
functions.

   outFile     the file to write
   byte        an array of bytes
   n           the number of bytes to write
   pcOrder     bytes are written in reverse order if TRUE
====================================================================== */

writeBytes : outFile, byte, n, pcOrder
{
   if ( pcOrder ) {
      first = n;
      stop  = 0;
      step  = -1;
   }
   else {
      first = 1;
      stop  = n + 1;
      step  = 1;
   }
   
   for ( i = first; i != stop; i += step )
      outFile.writeByte( byte[ i ] );
}

