// STL Importer for Lightwave 7.0+
//
//   Written By Sean Hyde-Moyer (Copywrite 2003)
//   This program is released to the public under the terms of the
//   GNU GENERAL PUBLIC LICENSE
//   See http://www.gnu.org/copyleft/gpl.txt
//   For the full text of the license agreement



@script modeler


main
{

	fType = 3;
	fDefault[1] = "Binary";
	fDefault[2] = "ASCII";
	fDefault[3] = "Auto";	

	pType = 1;
	pDefault[1] = "Intel";
	pDefault[2] = "Mac";


//______Prep File to Write_______________________________________________________________


    	directory = recall ("directory","Drive");



	reqbegin("SHM STL Importer");
    	reqsize(500,145);

    	YPOS = 10; 


		f1 = ctlfilename("Import STL File",directory,100,1);
        		ctlposition(f1,10,YPOS,450,20);
			YPOS = YPOS + 25;

 		t1 = ctlchoice("Type",fType,fDefault);
        		ctlposition(t1,71,YPOS,170,20);
			YPOS = YPOS + 25;

 		t2 = ctlchoice("Byte Order",pType,pDefault);
        		ctlposition(t2,45,YPOS,150,20);
			YPOS = YPOS + 25;

		s1 = ctlcheckbox("Fix Inverted", false);
        		ctlposition(s1,100,YPOS,150,20);

	if(reqpost())
	{
		stlFileName = getvalue(f1);
		readType = getvalue(t1);
		platformType = getvalue(t2);
		invertIt = getvalue(s1);
	}
	else
		return;
	reqend();

	isPC = TRUE;
	if(platformType == 2) isPC = FALSE;


	//_____________________Do Some Error Checking


	if(stlFileName == nil)
		return;


	base = split(stlFileName);

	stlname = base[3];





	//_____________________Auto Parse File Type

	if(readType == 3){	

		readType = 2;
		
		testFile = File(stlFileName,"rb");

			if(testFile == nil){
				error("Cannot Read File \"",stlFileName,"\"");
				return;
			}

		testFile.offset(84,FROMSTART);

		for(testAsc = 1; testAsc <= 20; ++testAsc)
		{
			testVal = testFile.readByte();
		
			if((testVal < 0) || (testVal > 127)){
				readType = 1;
			}
		}


	}


//______Open File__________________________________________

	if(readType == 1){
		stlFile = File(stlFileName,"rb");
	}
	else{
		stlFile = File(stlFileName,"r");
	}


	if(stlFile == nil){
		error("Cannot Read File \"",stlFileName,"\"");
		return;
	}

        store("directory",stlFileName);



//______Import as Binary STL_____________________________________________________________________


	if(readType == 1){

		//______Skip Header of 80 Bytes; Content is Irrelevant__________________________	

		for(headloop = 1; headloop <= 80; ++headloop)
		{
				junk = stlFile.readByte();				
		}	 



		selmode(USER);				
		selpolygon(CLEAR,NVGT,0);

		editbegin();




		//______Read Polycount (4 Byte Int)______________________________________________

	
		manypolys = readInt( stlFile, isPC );
		

		moninit(manypolys,"Importing Binary STL");
		
	
		//______Read Body of File________________________________________________________


		for(loop=1; loop <= manypolys; loop++)	
		{
		
			monstep();			

			//______Read Normal: 3 Floats__________________________________________



				for(normloop = 1; normloop <=3; ++normloop)
				{
						coord[normloop] = (readFloat( stlFile, isPC ));
				}			



				//______Write Vertices as 3 Floats________________________________
	
	
				for(vertloop = 1; vertloop <=3; ++vertloop)
				{

					for(axloop = 1; axloop <=3; ++axloop)
					{

						coord[axloop] = (readFloat( stlFile, isPC ));
						
					}

					point[vertloop] = addpoint(coord[1],coord[2],coord[3]);

				}


				//______Take 3 Verts & Make a Poly


				polygon = addpolygon(point);


				//______Write 2 Byte Pad (Could Contain Color Extension)__________________________________________
			
					
				stlFile.readByte();
				stlFile.readByte();			

		}


		editend();
	monend();

	stlFile.close();

	mergepoints(0);

	}



//____________Import as Ascii STL_______________________________________________________________


	if(readType == 2){

		fileSize = stlFile.linecount();

		manypolys = (fileSize - 2)/7;

		moninit(manypolys,"Importing ASCII STL");

			selmode(USER);				
			selpolygon(CLEAR,NVGT,0);

			editbegin();
	
			stlFile.line(2); //______Skip Header

			for(loop=1; loop <= manypolys; loop++)	
			{
			
				monstep();				

				aNormal = stlFile.read();

				outerLoop = stlFile.read();
				
				for(vertloop = 1; vertloop <= 3; ++vertloop)
				{
					getVert = stlFile.read();

					tempArray = parse(" ",getVert);

					coord[1] = number(tempArray[2]);
					coord[2] = number(tempArray[3]);
					coord[3] = number(tempArray[4]);

					point[vertloop] = addpoint(coord[1],coord[2],coord[3]);
				}

				polygon = addpolygon(point);

				endLoop = stlFile.read();
				endFacet = stlFile.read();

			}


			editend();

	mergepoints(0);

	monend();

	}

	if(invertIt == 1){
		scale(<-1,1,1>);
		flip();
	}



}





/*
======================================================================
Binary number read/write functions
Ernie Wright  15 Oct 03
====================================================================== */


/*
======================================================================
readShort()    2-byte signed two's-complement integers
readInt()      4-byte signed two's-complement integers
readFloat()    4-byte IEEE floating-point

Read a number from a binary file.  If 'pcOrder' is TRUE, the number is
read in little-endian (Windows) byte order, otherwise it's read in
big-endian (Motorola, Internet) order.
====================================================================== */

readShort : inFile, pcOrder
{
   var byte[ 2 ];
   
   readBytes( inFile, byte, 2, pcOrder );

   if ( byte[ 1 ] < 128 )
      return byte[ 1 ] * 256 + byte[ 2 ];
   else
      return -(( 255 - byte[ 1 ] ) * 256 + ( 256 - byte[ 2 ] ));
}

   
readInt : inFile, pcOrder
{
   var byte[ 4 ];
   
   readBytes( inFile, byte, 4, pcOrder );

   if ( byte[ 1 ] < 128 )      
      return byte[ 1 ] * 16777216
           + byte[ 2 ] * 65536
           + byte[ 3 ] * 256
           + byte[ 4 ];
   else
      return -(( 255 - byte[ 1 ] ) * 16777216
             + ( 255 - byte[ 2 ] ) * 65536
             + ( 255 - byte[ 3 ] ) * 256
             + ( 256 - byte[ 4 ] ));
}


readFloat : inFile, pcOrder
{
   var byte[ 4 ];
   
   readBytes( inFile, byte, 4, pcOrder );

   if ( byte[ 1 ] == 0 && byte[ 2 ] == 0 && byte[ 3 ] == 0 && byte[ 4 ] == 0 )
	return 0;

   m = 1.0 + (( byte[ 2 ] % 128 ) * 65536 + byte[ 3 ] * 256 + byte[ 4 ] )
     / 8388608.0;
   e = ( byte[ 1 ] % 128 ) * 2 + integer( byte[ 2 ] / 128 ) - 127;
   x = m * pow( 2.0, e );
   if ( byte[ 1 ] > 127 ) x = -x;

   return x;
}


/*
======================================================================
writeShort()    2-byte signed two's complement integers
writeInt()      4-byte signed two's complement integers
writeFloat()    4-byte IEEE floating-point

Write a number to a binary file.  If 'pcOrder' is TRUE, the number is
written in little-endian (Windows) byte order, otherwise it's written
in big-endian (Motorola, Internet) order.
====================================================================== */

writeShort : outFile, i, pcOrder
{
   var byte[ 2 ];

   if ( i >= 0 ) {
      byte[ 1 ] = integer( i / 256 ) % 256;
      byte[ 2 ] = i % 256;
   }
   else {
      j = -i - 1;
      byte[ 1 ] = 255 - ( integer( j / 256 ) % 256 );
      byte[ 2 ] = 255 - ( j % 256 );
   }
   
   writeBytes( outFile, byte, 2, pcOrder );
} 


writeInt : outFile, i, pcOrder
{
   var byte[ 4 ];

   if ( i >= 0 ) {
      byte[ 1 ] = integer( i / 16777216 );
      byte[ 2 ] = integer( i / 65536 ) % 256;
      byte[ 3 ] = integer( i / 256 ) % 256;
      byte[ 4 ] = i % 256;
   }
   else {
      j = -i - 1;
      byte[ 1 ] = 255 - integer( j / 16777216 );
      byte[ 2 ] = 255 - ( integer( j / 65536 ) % 256 );
      byte[ 3 ] = 255 - ( integer( j / 256 ) % 256 );
      byte[ 4 ] = 255 - ( j % 256 );
   }
   
   writeBytes( outFile, byte, 4, pcOrder );
}


writeFloat : outFile, x, pcOrder
{
   var byte[ 4 ];

   if(x==0)
   {
      	byte[ 4 ] = 0;
	byte[ 3 ] = 0;
	byte[ 2 ] = 0;
	byte[ 1 ] = 0;
   }
   else
   {
	   lgx = log( abs( x )) / log( 2.0 );
	   e = integer( floor( lgx ));
	   m = integer( floor(( pow( 2.0, lgx - e ) - 1.0 ) * 8388608 ));
	   e += 127;

	   byte[ 1 ] = integer( e / 2 ) + ( x < 0 ? 128 : 0 );
	   byte[ 2 ] = integer( m / 65536 ) + 128 * ( e % 2 );
	   byte[ 3 ] = integer( m / 256 ) % 256;
	   byte[ 4 ] = m % 256;
   }

   writeBytes( outFile, byte, 4, pcOrder );

}


/*
======================================================================
readBytes()

Read bytes from a file.  Called by the number reading functions.

   inFile      the file to read
   byte        array where bytes are stored
   n           the number of bytes to read from inFile
   pcOrder     bytes are stored in reverse order if TRUE
====================================================================== */

readBytes : inFile, byte, n, pcOrder
{
   if ( pcOrder ) {
      first = n;
      stop  = 0;
      step  = -1;
   }
   else {
      first = 1;
      stop  = n + 1;
      step  = 1;
   }
   
   for ( i = first; i != stop; i += step )
      byte[ i ] = inFile.readByte();
}


/*
======================================================================
writeBytes()

Write an array of bytes to a file.  Called by the number writing
functions.

   outFile     the file to write
   byte        an array of bytes
   n           the number of bytes to write
   pcOrder     bytes are written in reverse order if TRUE
====================================================================== */

writeBytes : outFile, byte, n, pcOrder
{
   if ( pcOrder ) {
      first = n;
      stop  = 0;
      step  = -1;
   }
   else {
      first = 1;
      stop  = n + 1;
      step  = 1;
   }
   
   for ( i = first; i != stop; i += step )
      outFile.writeByte( byte[ i ] );
}

